import requests
import typing as t
from .base import BaseProvider
from ..models import Domain
from ..types import InstrumentRecord, InstrumentInfo


class CloudflareProvider(BaseProvider):
    TIMEOUT = 60

    @property
    def base_url(self) -> str:
        zone_id = self.options["zone_id"]
        return f'https://api.cloudflare.com/client/v4/zones/{zone_id}/'

    @property
    def auth_key(self):
        return self.options["auth_key"]

    def skip_hostname(self, hostname: str) -> bool:
        skips = self.options.get('ignore_hostnames')
        if not skips:
            return False
        return hostname.endswith(tuple(skips))

    def request(self, method: str, path: str, **kwargs):
        url = self.base_url + path
        headers = {'Authorization': f'Bearer {self.auth_key}'}
        timeout = self.options.get("timeout", self.TIMEOUT)
        return requests.request(method, url, timeout=timeout, headers=headers, **kwargs)

    def _perform_domain_request(self, method, identifier=None, **kwargs):
        if identifier is None:
            path = 'custom_hostnames'
        else:
            path = f'custom_hostnames/{identifier}'
        resp = self.request(method, path, **kwargs)
        return resp.json()

    def add_domain(self, domain: Domain) -> Domain:
        if self.skip_hostname(domain.hostname):
            return domain

        ssl_settings = self.options.get('ssl_settings', {})
        if domain.instrument_id:
            payload = {'ssl': ssl_settings}
            resp = self._perform_domain_request(
                'PATCH',
                identifier=domain.instrument_id,
                json=payload,
            )
            update_domain(domain, resp)
        else:
            payload = {
                'hostname': domain.hostname,
                'ssl': ssl_settings,
            }
            resp = self._perform_domain_request('POST', json=payload)
            update_domain(domain, resp)
        return domain

    def verify_domain(self, domain: Domain) -> Domain:
        if self.skip_hostname(domain.hostname):
            return domain

        if domain.instrument_id:
            resp = self._perform_domain_request(
                'GET',
                identifier=domain.instrument_id,
            )
            update_domain(domain, resp)
        return domain

    def remove_domain(self, domain: Domain) -> None:
        if domain.instrument_id and not self.skip_hostname(domain.hostname):
            self._perform_domain_request('DELETE', identifier=domain.instrument_id)


def normalize_instrument(resp) -> t.Tuple[str, InstrumentInfo]:
    errors = []
    if 'errors' in resp:
        for item in resp['errors']:
            errors.append(item['message'])

    if 'verification_errors' in resp:
        errors.extend(resp['verification_errors'])

    messages = []
    if 'messages' in resp:
        for item in resp['messages']:
            messages.append(item['message'])

    result = resp.get("result")
    instrument: InstrumentInfo = {
        "ownership_status": result["status"],
    }
    records: t.List[InstrumentRecord] = []
    ownership_verification = result.get("ownership_verification")
    if ownership_verification:
        records.append(ownership_verification)

    ssl_config = result.get("ssl")
    if ssl_config:
        instrument["ssl_status"] = ssl_config["status"]
        validation_records = ssl_config.get("validation_records")
        if validation_records:
            for item in validation_records:
                name = item["txt_name"]
                value = item["txt_value"]
                records.append({"name": name, "value": value, "type": "txt"})

        validation_errors = ssl_config.get("validation_errors")
        if validation_errors:
            for item in validation_errors:
                errors.append(item['message'])

    instrument["records"] = records
    return result["id"], instrument


def update_domain(obj: Domain, resp):
    instrument_id, instrument = normalize_instrument(resp)

    if instrument.get("ownership_status") == "active":
        obj.verified = True

    if instrument.get("ssl_status") == "active":
        obj.ssl = True

    obj.instrument_id = instrument_id
    obj.instrument = instrument
    obj.active = obj.verified and obj.ssl
    obj.save()



import typing as t


InstrumentRecord = t.TypedDict("InstrumentRecord", {
    "name": str,
    "type": t.Literal["txt", "cname"],
    "value": str,
})
InstrumentInfo = t.TypedDict("InstrumentInfo", {
    "errors": t.List[str],
    "messages": t.List[str],
    "ownership_status": t.Literal["active", "pending", "failed"],
    "ssl_status": t.Literal["active", "pending", "failed"],
    "records": t.List[InstrumentRecord],
}, total=False)



from saas_base.test import SaasTestCase
from saas_domain.models import Domain

TEST_DATA = {
    "hostname": "example.com",
    "provider": "null",
}


class TestDomainAPIWithOwner(SaasTestCase):
    user_id = SaasTestCase.OWNER_USER_ID

    def test_list_domains(self):
        self.force_login()

        resp = self.client.get('/m/domains/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json(), [])

        Domain.objects.create(tenant=self.tenant, hostname="example.com")
        resp = self.client.get('/m/domains/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()), 1)
        self.assertEqual(resp.json()[0]['hostname'], 'example.com')

    def test_create_domain(self):
        self.force_login()
        resp = self.client.post('/m/domains/', data=TEST_DATA)
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp.json()["hostname"], "example.com")
        self.assertEqual(resp.json()["verified"], False)

    def test_create_domain_with_invalid_provider(self):
        self.force_login()
        payload = {
            "hostname": "example.com",
            "provider": "invalid",
        }
        resp = self.client.post('/m/domains/', data=payload)
        self.assertEqual(resp.status_code, 400)


class TestDomainAPIWithGuestUser(SaasTestCase):
    user_id = SaasTestCase.GUEST_USER_ID

    def test_list_domains_with_read_permission(self):
        self.add_user_perms('tenant.read')
        self.force_login()

        resp = self.client.get('/m/domains/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json(), [])

        Domain.objects.create(tenant=self.tenant, hostname="example.com")
        resp = self.client.get('/m/domains/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()), 1)
        self.assertEqual(resp.json()[0]['hostname'], 'example.com')

    def test_list_domains_without_permission(self):
        self.force_login()
        resp = self.client.get('/m/domains/')
        self.assertEqual(resp.status_code, 403)

    def test_create_domain_with_admin_permission(self):
        self.add_user_perms('tenant.admin')
        self.force_login()
        resp = self.client.post('/m/domains/', data=TEST_DATA)
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp.json()["hostname"], "example.com")
        self.assertEqual(resp.json()["verified"], False)

    def test_create_domain_with_read_permission(self):
        resp = self.client.post('/m/domains/', data=TEST_DATA)
        self.assertEqual(resp.status_code, 403)

    def test_retrieve_domain_with_read_permission(self):
        self.add_user_perms('tenant.read')
        self.force_login()

        domain = Domain.objects.create(
            tenant=self.tenant,
            hostname="example.com",
            provider="null",
        )
        resp = self.client.get(f'/m/domains/{domain.pk}/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()["verified"], False)

        resp = self.client.get(f'/m/domains/{domain.pk}/?verify=true')
        self.assertEqual(resp.json()["verified"], True)

    def test_enable_and_refresh_domain(self):
        self.add_user_perms('tenant.admin')
        self.force_login()
        domain = Domain.objects.create(
            tenant=self.tenant,
            hostname="example.com",
            provider="null",
        )

        # enable domain
        resp = self.client.post(f'/m/domains/{domain.pk}/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()["instrument"]['ownership_status'], 'pending')

    def test_delete_domain_with_admin_permission(self):
        self.add_user_perms('tenant.admin')
        self.force_login()

        domain = Domain.objects.create(tenant=self.tenant, hostname="example.com")
        resp = self.client.delete(f'/m/domains/{domain.pk}/')
        self.assertEqual(resp.status_code, 204)



from django.urls import path
from ..endpoints.domain import DomainListEndpoint, DomainItemEndpoint

urlpatterns = [
    path('', DomainListEndpoint.as_view()),
    path('<pk>/', DomainItemEndpoint.as_view()),
]



import logging
from .base import BaseProvider
from ..models import Domain

logger = logging.getLogger(__name__)


class NullProvider(BaseProvider):
    def add_domain(self, domain: Domain) -> Domain:
        logger.info("Add domain:", domain)
        domain.instrument_id = 'null'
        domain.instrument = {
            'ownership_status': 'pending',
            'ssl_status': 'pending',
            'records': [
                {'name': '@', 'type': 'cname', 'value': 'testserver'},
            ],
        }
        domain.save()
        return domain

    def verify_domain(self, domain: Domain) -> Domain:
        logger.info("Verify domain:", domain)
        domain.active = True
        domain.ssl = True
        domain.verified = True
        domain.save()
        return domain

    def remove_domain(self, domain: Domain):
        logger.info("Remove domain:", domain)



from unittest.mock import patch
from saas_domain.models import Domain, DomainManager
from saas_base.test import SaasTestCase


class TestDomainManager(SaasTestCase):
    def test_not_found(self):
        with patch.object(DomainManager, "get", side_effect=Domain.DoesNotExist()) as mock_get:
            value = Domain.objects.get_tenant_id('google.com')
            self.assertIsNone(value)

            # trigger get from cache
            value = Domain.objects.get_tenant_id('google.com')
            self.assertIsNone(value)
            self.assertEqual(mock_get.call_count, 1)

    def test_found(self):
        tenant = self.get_tenant()
        Domain.objects.create(tenant=tenant, hostname='example.com')

        with patch.object(DomainManager, "get", side_effect=Domain.DoesNotExist()) as mock_get:
            value = Domain.objects.get_tenant_id('example.com')
            self.assertEqual(value, tenant.pk)

            value = Domain.objects.get_tenant_id('example.com')
            self.assertEqual(value, tenant.pk)
            mock_get.assert_not_called()

    def test_purge_tenant_events(self):
        from saas_domain.providers import NullProvider

        tenant = self.get_tenant()
        Domain.objects.create(tenant=tenant, hostname='example.com', provider='null')

        with patch.object(NullProvider, "remove_domain") as remove_domain:
            tenant.delete()
            remove_domain.assert_called_once()



import typing as t
import uuid
from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from saas_base.db import CachedManager

__all__ = ["Domain", "DomainManager"]


class DomainManager(CachedManager["Domain"]):
    natural_key = ["hostname"]

    def get_by_natural_key(self, hostname: str):
        return self.get_from_cache_by_natural_key(hostname)

    def get_tenant_id(self, hostname: str) -> t.Optional[int]:
        try:
            instance = self.get_by_natural_key(hostname)
            return instance.tenant_id
        except self.model.DoesNotExist:
            return None


class Domain(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)
    tenant = models.ForeignKey(settings.SAAS_TENANT_MODEL, on_delete=models.CASCADE)
    provider = models.CharField(max_length=100)
    hostname = models.CharField(max_length=100, unique=True)
    verified = models.BooleanField(default=False, editable=False)
    ssl = models.BooleanField(default=False, editable=False)
    active = models.BooleanField(default=False, editable=False)
    instrument_id = models.CharField(max_length=256, null=True, blank=True, editable=False)
    instrument = models.JSONField(blank=True, null=True, editable=False)
    created_at = models.DateTimeField(default=timezone.now, editable=False)
    updated_at = models.DateTimeField(auto_now=True, editable=False)
    objects = DomainManager()

    class Meta:
        verbose_name = _("domain")
        verbose_name_plural = _("domains")
        ordering = ['created_at']
        db_table = 'saas_domain'

    def __str__(self):
        return self.hostname

    @property
    def base_url(self) -> str:
        if self.ssl:
            return f'https://{self.hostname}'
        return f'http://{self.hostname}'

    def natural_key(self):
        return self.hostname,

    def disable(self):
        self.verified = False
        self.active = False
        self.ssl = False
        self.instrument = None
        self.save()



from rest_framework import serializers
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, OpenApiExample, extend_schema
from drf_spectacular.extensions import OpenApiViewExtension


class DomainListEndpoint(OpenApiViewExtension):
    target_class = 'saas_domain.endpoints.domain.DomainListEndpoint'

    def view_replacement(self):
        class Fixed(self.target_class):
            @extend_schema(
                operation_id='domains_list',
                summary='List Domains',
            )
            def get(self, *args, **kwargs):
                pass

            @extend_schema(
                operation_id='domains_add',
                summary='Add Domain',
            )
            def post(self, *args, **kwargs):
                pass
        return Fixed



class DomainItemEndpoint(OpenApiViewExtension):
    target_class = 'saas_domain.endpoints.domain.DomainItemEndpoint'

    def view_replacement(self):
        class Fixed(self.target_class):
            @extend_schema(
                operation_id='domains_retrieve',
                summary='Retrieve Domain',
            )
            def get(self, *args, **kwargs):
                pass

            @extend_schema(
                operation_id='domains_update',
                summary='Re-add Domain',
            )
            def post(self, *args, **kwargs):
                pass

            @extend_schema(
                operation_id='domains_destroy',
                summary='Remove Domain',
            )
            def delete(self, *args, **kwargs):
                pass
        return Fixed



from .client import FixturesTestCase

TEST_DATA = {
    "hostname": "example.com",
    "provider": "cloudflare",
}


class TestCloudflareAPI(FixturesTestCase):
    user_id = FixturesTestCase.OWNER_USER_ID

    def test_add_domain(self):
        self.force_login()
        with self.mock_requests('cloudflare_create_success.json'):
            resp = self.client.post('/m/domains/', data=TEST_DATA)
            self.assertEqual(resp.status_code, 201)
            self.assertEqual(resp.json()["hostname"], "example.com")
            self.assertEqual(resp.json()["verified"], True)
            self.assertEqual(resp.json()["ssl"], False)

    def test_add_ignore_hostnames(self):
        self.force_login()
        with self.mock_requests('cloudflare_create_success.json'):
            resp = self.client.post('/m/domains/', data={
                "hostname": "blog.localtest.me",
                "provider": "cloudflare",
            })
            self.assertEqual(resp.status_code, 201)
            self.assertEqual(resp.json()["hostname"], "blog.localtest.me")
            self.assertEqual(resp.json()["verified"], False)
            self.assertEqual(resp.json()["ssl"], False)

    def test_verify_domain(self):
        self.force_login()
        with self.mock_requests(
            'cloudflare_create_success.json',
            'cloudflare_verify_success.json',
        ):
            resp = self.client.post('/m/domains/', data=TEST_DATA)
            self.assertEqual(resp.status_code, 201)
            data = resp.json()
            domain_id = data["id"]
            resp = self.client.get(f'/m/domains/{domain_id}/?verify=true')
            self.assertEqual(resp.status_code, 200)

    def test_remove_domain(self):
        self.force_login()
        with self.mock_requests(
            'cloudflare_create_success.json',
            'cloudflare_delete_success.json',
        ):
            resp = self.client.post('/m/domains/', data=TEST_DATA)
            self.assertEqual(resp.status_code, 201)
            data = resp.json()
            domain_id = data["id"]
            resp = self.client.delete(f'/m/domains/{domain_id}/')
            self.assertEqual(resp.status_code, 204)



from django.core.signals import setting_changed
from saas_base.settings import Settings

DEFAULTS = {
    'SUPPORTED_PROVIDERS': [],
    'BLOCKED_DOMAINS': [],
    'PROVIDERS': {},
}


class DomainSettings(Settings):
    IMPORT_PROVIDERS = [
        'PROVIDERS',
    ]

    def get_supported_providers(self):
        if self.SUPPORTED_PROVIDERS:
            return self.SUPPORTED_PROVIDERS
        return list(self.PROVIDERS.keys())


domain_settings = DomainSettings('SAAS_DOMAIN', defaults=DEFAULTS)
setting_changed.connect(domain_settings.listen_setting_changed)



from django.http.request import split_domain_port
from saas_base.models import get_tenant_model
from .models import Domain

TenantModel = get_tenant_model()


__all__ = [
    'DomainTenantIdMiddleware',
]


class DomainTenantIdMiddleware:
    def __init__(self, get_response=None):
        self.get_response = get_response

    @staticmethod
    def get_tenant_id(request):
        tenant_id = getattr(request, '_cached_tenant_id', None)
        if tenant_id:
            return tenant_id

        host = request.get_host()
        hostname = split_domain_port(host)[0]
        return Domain.objects.get_tenant_id(hostname)

    def __call__(self, request):
        request.tenant_id = self.get_tenant_id(request)
        return self.get_response(request)



from django.contrib import admin
from .models import Domain


@admin.register(Domain)
class DomainAdmin(admin.ModelAdmin):
    list_display = ('id', 'tenant', 'hostname', 'provider', 'active', 'created_at')



from django.urls import path, include
from drf_spectacular.views import SpectacularJSONAPIView

urlpatterns = [
    path('m/domains/', include('saas_domain.api_urls.domain')),
    path('schema/openapi', SpectacularJSONAPIView.as_view(), name='schema'),
]



from django.apps import AppConfig


class DomainConfig(AppConfig):
    name = "saas_domain"
    verbose_name = "SaaS Domain"

    def ready(self):
        from ._ready import register_checks
        register_checks()
